@@ -14,6 +14,10 @@
  *  $Id$
  */
 
+#ifdef HAVE_CONFIG_H
+#include "config.h"
+#endif
+
 #include <stdlib.h>
 #include <stdio.h>
 #include <string.h>
@@ -50,10 +54,13 @@ lo_server_thread lo_server_thread_new(const char *port,
 
 static lo_server_thread alloc_server_thread(lo_server s)
 {
+    lo_server_thread st;
+
     if (!s)
         return NULL;
-    lo_server_thread st = (lo_server_thread)
-        malloc(sizeof(struct _lo_server_thread));
+
+    st = (lo_server_thread) malloc(sizeof(struct _lo_server_thread));
+
     st->s = s;
     st->active = 0;
     st->done = 0;
@@ -72,6 +79,15 @@ lo_server_thread lo_server_thread_new_multicast(const char *group,
     return st;
 }
 
+lo_server_thread lo_server_thread_new_multicast_iface(const char *group, const char *port,
+						      const char *iface, const char *ip,
+						      lo_err_handler err_h)
+{
+    lo_server_thread st = alloc_server_thread(
+        lo_server_new_multicast_iface(group, port, iface, ip, err_h));
+    return st;
+}
+
 lo_server_thread lo_server_thread_new_with_proto(const char *port,
                                                  int proto,
                                                  lo_err_handler err_h)
@@ -274,6 +290,7 @@ static void* thread_func(void *data)
 #error "No threading implementation selected."
 #endif
 #endif
+    return 0;
 }
 
 void lo_server_thread_pp(lo_server_thread st)
